import gzip
import json
from enum import Enum

from selenium.common.exceptions import (ElementNotInteractableException,
                                        NoSuchElementException,
                                        TimeoutException)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

import chase.urls as urls

from .session import ChaseSession


class PriceType(str, Enum):
    """This is an :class: 'enum.Enum' that contains the valid price types for an order."""

    LIMIT = "LIMIT"
    MARKET = "MARKET"
    STOP = "STOP"
    STOP_LIMIT = "STOP_LIMIT"


class Duration(str, Enum):
    """This is an :class:'~enum.Enum' that contains the valid durations for an order."""

    DAY = "DAY"
    GTC = "GOOD_TILL_CANCELLED"
    ON_OPEN = "ON_THE_OPEN"
    ON_CLOSE = "ON_THE_CLOSE"
    I_C = "IMMEDIATE_OR_CANCEL"


class OrderType(str, Enum):
    """
    This is an :class:'~enum.Enum'
    that contains the valid order types for an order.
    """

    BUY = "BUY"
    SELL = "SELL"
    SELL_ALL = "SELL_ALL"


class TypeCode(str, Enum):
    """
    This is an :class:'~enum.Enum'
    that contains the valid order types for an order.
    """

    CASH = "CASH"
    MARGIN = "MARGIN"


class Order:
    """
    This class contains information about an order.
    It also contains a method to place an order.
    """

    def __init__(self, session: ChaseSession, accept_warning: bool = True):
        self.session = session
        self.accept_warning = accept_warning
        self.order_number: str = ""

    def place_order(
        self,
        account_id,
        quantity,
        price_type: PriceType,
        symbol,
        duration: Duration,
        order_type: OrderType,
        limit_price: float = 0.00,
        stop_price: float = 0.00,
        after_hours: bool = True,
        dry_run=True,
    ):
        """
        Builds and places an order.
        :attr: 'order_confirmation`
        contains the order confirmation data after order placement.

        Args:
            account (str): Account number of the account to place the order in.
            symbol (str): Ticker to place the order for.
            order_type (PriceType): Price Type i.e. LIMIT, MARKET, STOP, etc.
            quantity (float): The number of shares to buy.
            duration (Duration): Duration of the order i.e. DAY, GT90, etc.
            price (float, optional): The price to buy the shares at. Defaults to 0.00.
            dry_run (bool, optional): Whether you want the order to be placed or not.
                                      Defaults to True.

        Returns:
            Order:order_confirmation: Dictionary containing the order confirmation data.
        """
        order_messages = {
            "ORDER INVALID": "",
            "WARNING": "",
            "ORDER PREVIEW": "",
            "AFTER HOURS WARNING": "",
            "ORDER CONFIRMATION": "",
        }

        for i in range(0, 4):
            self.session.driver.get(urls.order_page(account_id))
            self.session.driver.refresh()
            try:
                WebDriverWait(self.session.driver, 20).until(
                    EC.presence_of_element_located((By.XPATH, "//label[text()='Buy']"))
                )

                quote_box = self.session.driver.find_element(
                    By.CSS_SELECTOR,
                    "#equitySymbolLookup-block-autocomplete-validate-input-field",
                )
                quote_box.clear()
                quote_box.send_keys(symbol)
                quote_box.send_keys(Keys.ENTER)

                WebDriverWait(self.session.driver, 10).until(
                    EC.presence_of_element_located((By.CSS_SELECTOR, ".NOTE"))
                )
                WebDriverWait(self.session.driver, 10).until(
                    EC.invisibility_of_element((By.ID, "default-spinner_2"))
                )
                order_messages["ORDER INVALID"] = "Order page loaded correctly."
                break
            except (TimeoutException, NoSuchElementException):
                order_messages[
                    "ORDER INVALID"
                ] = f"Order page did not load correctly cannot continue. Tried {i + 1} times."
                print(order_messages["ORDER INVALID"])

        if order_messages["ORDER INVALID"] != "Order page loaded correctly.":
            return order_messages

        if order_type == "BUY":
            WebDriverWait(self.session.driver, 60).until(
                EC.element_to_be_clickable((By.XPATH, "//label[text()='Buy']"))
            ).click()
        elif order_type == "SELL":
            self.session.driver.find_element(By.XPATH, "//label[text()='Sell']").click()
        elif order_type == "SELL_ALL":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Sell All']"
            ).click()

        if price_type == "LIMIT":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Limit']"
            ).click()
        elif price_type == "MARKET":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Market']"
            ).click()
            if duration not in ["DAY", "ON_THE_CLOSE"]:
                order_messages[
                    "ORDER INVALID"
                ] = "Market orders must be DAY or ON THE CLOSE."
                return order_messages
        elif price_type == "STOP":
            self.session.driver.find_element(By.XPATH, "//label[text()='Stop']").click()
            if duration not in ["DAY", "GOOD_TILL_CANCELLED"]:
                order_messages[
                    "ORDER INVALID"
                ] = "Stop orders must be DAY or GOOD TILL CANCELLED."
                return order_messages
        elif price_type == "STOP_LIMIT":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Stop Limit']"
            ).click()
            if duration not in ["DAY", "GOOD_TILL_CANCELLED"]:
                order_messages[
                    "ORDER INVALID"
                ] = "Stop orders must be DAY or GOOD TILL CANCELLED."
                return order_messages

        if price_type in ["LIMIT", "STOP", "STOP_LIMIT"]:
            WebDriverWait(self.session.driver, 60).until(
                EC.presence_of_element_located((By.NAME, "tradeLimitPrice"))
            ).send_keys(limit_price)
        if price_type in ["STOP", "STOP_LIMIT"]:
            WebDriverWait(self.session.driver, 60).until(
                EC.presence_of_element_located((By.NAME, "tradeStopPrice"))
            ).send_keys(stop_price)

        quantity_box = self.session.driver.find_element(By.NAME, "tradeQuantity")
        quantity_box.clear()
        quantity_box.send_keys(quantity)

        if duration == "DAY":
            self.session.driver.find_element(By.XPATH, "//label[text()='Day']").click()
        elif duration == "GOOD_TILL_CANCELLED":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Good 'til canceled']"
            ).click()
        elif duration == "ON_THE_OPEN":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='On open']"
            ).click()
        elif duration == "ON_THE_CLOSE":
            self.session.driver.find_element(
                By.XPATH, "//label[text()='On close']"
            ).click()
        elif duration == "IMMEDIATE_OR_CANCEL":
            self.session.driver.find_element(
                By.CSS_SELECTOR, "#tradeExecutionOptions-iconwrap"
            ).click()
            self.session.driver.find_element(
                By.XPATH, "//label[text()='Immediate or Cancel']"
            ).click()

        try:
            WebDriverWait(self.session.driver, 30).until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, "#previewOrder"))
            ).click()
        except NoSuchElementException:
            raise Exception("No preview button found. Cannot continue.")
        except ElementNotInteractableException:
            raise Exception("Preview button not interactable. Cannot continue.")

        try:
            warning = (
                WebDriverWait(self.session.driver, 5)
                .until(
                    EC.presence_of_element_located(
                        (
                            By.CSS_SELECTOR,
                            "#entry-trade-wrapper > div > div:nth-child(1) > div > div",
                        )
                    )
                )
                .text
            )
            order_messages["ORDER INVALID"] = warning
            return order_messages
        except (NoSuchElementException, TimeoutException):
            order_messages["ORDER INVALID"] = "No invalid order message found."

        try:
            WebDriverWait(self.session.driver, 5).until(
                EC.url_to_be(urls.warning_page(account_id))
            )
            warning = self.session.driver.find_element(
                By.CSS_SELECTOR, ".singleWarning"
            ).text
            order_messages["WARNING"] = warning
            if self.accept_warning:
                try:
                    WebDriverWait(self.session.driver, 5).until(
                        EC.presence_of_element_located(
                            (By.CSS_SELECTOR, "#acceptWarnings")
                        )
                    ).click()
                except (NoSuchElementException, TimeoutException):
                    raise Exception("No accept button found. Could not dismiss prompt.")
            else:
                return order_messages
        except TimeoutException:
            order_messages["WARNING"] = "No warning page found."

        try:
            WebDriverWait(self.session.driver, 5).until(
                EC.url_to_be(urls.order_preview_page(account_id))
            )
            order_preview = self.session.driver.find_element(
                By.CLASS_NAME, "trade-wrapper"
            ).text
            order_messages["ORDER PREVIEW"] = order_preview
            if not dry_run:
                try:
                    self.session.driver.find_element(
                        By.CSS_SELECTOR, "#submitOrder"
                    ).click()
                except NoSuchElementException:
                    raise Exception("No place order button found cannot continue.")
            else:
                return order_messages
        except TimeoutException:
            order_messages["ORDER PREVIEW"] = "No order preview page found."

        try:
            WebDriverWait(self.session.driver, 5).until(
                EC.url_to_be(urls.after_hours_warning(account_id))
            )
            warning = self.session.driver.find_element(
                By.CSS_SELECTOR, "#afterHoursModal > div.markets-message > div"
            ).text
            order_messages["AFTER HOURS WARNING"] = warning
            if after_hours:
                try:
                    self.session.driver.find_element(
                        By.CSS_SELECTOR, "#confirmAfterHoursOrder"
                    ).click()
                except NoSuchElementException:
                    raise Exception("No yes button found. Could not dismiss prompt.")
            else:
                return order_messages
        except TimeoutException:
            order_messages["AFTER HOURS WARNING"] = "No after hours warning page found."

        try:
            WebDriverWait(self.session.driver, 5).until(
                EC.url_to_be(urls.order_confirmation())
            )
            order_confirmation = self.session.driver.find_element(
                By.ID, "confirmationHeader"
            ).text
            order_confirmation = (
                order_confirmation
                + " "
                + self.session.driver.find_element(By.ID, "confirmationAlert").text
            )
            order_confirmation = (
                order_confirmation
                + " "
                + self.session.driver.find_element(By.CLASS_NAME, "trade-wrapper").text
            )
            order_confirmation = order_confirmation.replace("\n", " ")
            order_messages["ORDER CONFIRMATION"] = order_confirmation
            return order_messages
        except TimeoutException:
            order_messages[
                "ORDER CONFIRMATION"
            ] = "No order confirmation page found. Order Failed."
            return order_messages

    def get_order_statuses(self, account_id):
        """
        Retrieves the statuses of all recent orders placed.

        This method navigates to the order status page and scrapes the status of each order.
        It returns a list of dictionaries, where each dictionary represents an order and contains
        the order number and status.

        Returns:
            list[dict]: A list of dictionaries, where each dictionary contains 'order_number' and 'status' keys.

        Raises:
            TimeoutException: If the order status page fails to load within the specified timeout.
            NoSuchElementException: If an expected element on the order status page cannot be found.
        """
        try:
            self.session.driver.get(urls.order_status(account_id))
            WebDriverWait(self.session.driver, 10).until(
                EC.presence_of_all_elements_located(
                    (By.CSS_SELECTOR, "tr.investmentGridRow")
                )
            )
            for request in self.session.driver.requests:
                if request.response and request.url == urls.order_info():
                    body = request.response.body
                    try:
                        body = body.decode("utf-8")
                        body = json.loads(body)
                    except UnicodeDecodeError:
                        order_json = gzip.decompress(body).decode("utf-8")
                        order_json = json.loads(order_json)
                        return order_json
            return body
        except (TimeoutException, NoSuchElementException, UnboundLocalError):
            return None



import gzip
import json
from datetime import datetime
from time import sleep

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .session import ChaseSession
from .urls import account_holdings, holdings_json, order_page, quote_endpoint


class SymbolQuote:
    """
    A class to manage the quote of a specific symbol associated with a ChaseSession.

    This class provides methods to retrieve and manage information about the quote of a specific symbol associated with a given session.

    Attributes:
        account_id (str): The ID of the account.
        session (ChaseSession): The session associated with the account.
        symbol (str): The symbol for which the quote is retrieved.
        ask_price (float): The ask price of the symbol.
        ask_exchange_code (str): The exchange code of the ask price.
        ask_quantity (int): The quantity of the ask price.
        bid_price (float): The bid price of the symbol.
        bid_exchange_code (str): The exchange code of the bid price.
        bid_quantity (int): The quantity of the bid price.
        change_amount (float): The change amount of the symbol.
        last_trade_price (float): The last trade price of the symbol.
        last_trade_quantity (int): The last trade quantity of the symbol.
        last_exchange_code (str): The exchange code of the last trade.
        change_percentage (float): The change percentage of the symbol.
        as_of_time (datetime): The timestamp of the quote information.
        security_description (str): The description of the security.
        security_symbol (str): The symbol of the security.
        raw_json (dict): The raw JSON response containing the quote information.

    Methods:
        get_symbol_quote(): Retrieves and sets the quote information of the symbol.
    """

    def __init__(self, account_id, session: ChaseSession, symbol: str):
        """
        Initializes a SymbolQuote object with a given account ID, ChaseSession, and symbol.

        Args:
            account_id (str): The ID of the account.
            session (ChaseSession): The session associated with the account.
            symbol (str): The symbol for which the quote is retrieved.
        """
        self.account_id = account_id
        self.session = session
        self.symbol = symbol
        self.ask_price: float = 0
        self.ask_exchange_code: str = ""
        self.ask_quantity: int = 0
        self.bid_price: float = 0
        self.bid_exchange_code: str = ""
        self.bid_quantity: int = 0
        self.change_amount: float = 0
        self.last_trade_price: float = 0
        self.last_trade_quantity: int = 0
        self.last_exchange_code: str = ""
        self.change_percentage: float = 0
        self.as_of_time: datetime = None
        self.security_description: str = ""
        self.security_symbol: str = ""
        self.raw_json: dict = {}
        self.get_symbol_quote()

    def get_symbol_quote(self):
        """
        Retrieves and sets the quote information of the symbol.

        This method navigates to the symbol quote page, waits for the quote information to load, and then retrieves the quote information from the page.

        Returns:
            None
        """
        self.session.driver.get(order_page(self.account_id))
        WebDriverWait(self.session.driver, 60).until(EC.presence_of_element_located((By.XPATH, "//label[text()='Buy']")))
        quote_box = self.session.driver.find_element(By.CSS_SELECTOR, "#equitySymbolLookup-block-autocomplete-validate-input-field")
        quote_box.send_keys(self.symbol)
        quote_box.send_keys(Keys.ENTER)
        WebDriverWait(self.session.driver, 60).until(EC.presence_of_element_located((By.CSS_SELECTOR, ".NOTE")))
        for request in self.session.driver.requests:
            if request.response and request.url == quote_endpoint(self.symbol):
                body = request.response.body
                body = gzip.decompress(body).decode('utf-8')
                self.raw_json = json.loads(body)
        self.ask_price = float(self.raw_json['askPriceAmount'])
        self.ask_exchange_code = self.raw_json['askExchangeCode']
        self.ask_quantity = int(self.raw_json['askQuantity'])
        self.bid_price = float(self.raw_json['bidPriceAmount'])
        self.bid_exchange_code = self.raw_json['bidExchangeCode']
        self.bid_quantity = int(self.raw_json['bidQuantity'])
        self.change_amount = float(self.raw_json['changeAmount'])
        self.last_trade_price = float(self.raw_json['lastTradePriceAmount'])
        self.last_trade_quantity = int(self.raw_json['lastTradeQuantity'])
        self.last_exchange_code = self.raw_json['lastTradeExchangeCode']
        self.change_percentage = float(self.raw_json['changePercent'])
        self.as_of_time = datetime.strptime(self.raw_json['asOfTimestamp'], "%Y-%m-%dT%H:%M:%S.%fZ")
        self.security_description = self.raw_json['securityDescriptionText']
        self.security_symbol = self.raw_json['securitySymbolCode']


class SymbolHoldings:
    """
    A class to manage the holdings of a specific account associated with a ChaseSession.

    This class provides methods to retrieve and manage information about the holdings of a specific account associated with a given session.

    Attributes:
        account_id (str): The ID of the account.
        session (ChaseSession): The session associated with the account.
        as_of_time (datetime): The timestamp of the holdings information.
        asset_allocation_tool_eligible_indicator (bool): Whether the account is eligible for the asset allocation tool.
        cash_sweep_position_summary (dict): The summary of the cash sweep position.
        custom_position_allowed_indicator (bool): Whether custom positions are allowed.
        error_responses (list): Any error responses returned when retrieving the holdings information.
        performance_allowed_indicator (bool): Whether performance tracking is allowed.
        positions (list): The positions held in the account.
        positions_summary (dict): The summary of the positions.
        raw_json (dict): The raw JSON response containing the holdings information.

    Methods:
        get_holdings(): Retrieves and sets the holdings information of the account.
    """

    def __init__(self, account_id, session: ChaseSession):
        """
        Initializes a SymbolHoldings object with a given account ID and ChaseSession.

        Args:
            account_id (str): The ID of the account.
            session (ChaseSession): The session associated with the account.
        """
        self.account_id = account_id
        self.session = session
        self.as_of_time: datetime = None
        self.asset_allocation_tool_eligible_indicator: bool = None
        self.cash_sweep_position_summary: dict = {}
        self.custom_position_allowed_indicator: bool = None
        self.error_responses: list = []
        self.performance_allowed_indicator: bool = None
        self.positions: list = []
        self.positions_summary: dict = {}
        self.raw_json: dict = {}

    def get_holdings(self):
        """
        Retrieves and sets the holdings information of the account.

        This method navigates to the account holdings page, waits for the holdings information to load, and then retrieves the holdings information from the page.

        Returns:
            bool: True if the holdings information was successfully retrieved, False otherwise.
        """
        self.session.driver.get(account_holdings(self.account_id))
        try:
            WebDriverWait(self.session.driver, 60).until(EC.presence_of_element_located((By.XPATH, "//*[@id='positions-tabs']")))
            sleep(5)
            for request in self.session.driver.requests:
                if request.response and request.url == holdings_json():
                    body = request.response.body
                    body = gzip.decompress(body).decode('utf-8')
                    self.raw_json = json.loads(body)
            self.as_of_time = datetime.strptime(self.raw_json['asOfTimestamp'], "%Y-%m-%dT%H:%M:%S.%fZ")
            self.asset_allocation_tool_eligible_indicator = bool(self.raw_json['assetAllocationToolEligibleIndicator'])
            self.cash_sweep_position_summary = self.raw_json['cashSweepPositionSummary']
            self.custom_position_allowed_indicator = bool(self.raw_json['customPositionAllowedIndicator'])
            self.error_responses = self.raw_json['errorResponses']
            self.performance_allowed_indicator = bool(self.raw_json['performanceAllowedIndicator'])
            self.positions = self.raw_json['positions']
            self.positions_summary = self.raw_json['positionsSummary']
            return True
        except (NoSuchElementException, TimeoutException):
            return False



import os
import traceback
from time import sleep

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.chrome.service import Service as ChromiumService
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from seleniumwire import webdriver
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.os_manager import ChromeType

from .urls import auth_code_page, login_page


class FileChange(FileSystemEventHandler):
    def __init__(self, filename):
        self.filename = filename
        self.file_modified = False

    def on_modified(self, event):
        if not event.is_directory and event.src_path.endswith(self.filename):
            self.file_modified = True

class ChaseSession:
    """
    A class to manage a session with Chase.

    This class provides methods to initialize a WebDriver with the necessary options, log into Chase, and perform other actions on the Chase website.

    Attributes:
        persistent_session (bool): Whether the session should be persistent across multiple uses of the WebDriver.
        headless (bool): Whether the WebDriver should run in headless mode.
        docker (bool): Whether the session is running in a Docker container.
        profile_path (str): The path to the user profile directory for the WebDriver.
        driver (selenium.webdriver.Chrome): The WebDriver instance used to interact with the Chase website.

    Methods:
        get_driver(): Initializes and returns a WebDriver with the necessary options.
        login(username, password, last_four): Logs into Chase with the provided credentials.
    """

    def __init__(self, title=None, headless=True, docker=False, external_code=False):
        """
        Initializes a new instance of the ChaseSession class.

        Args:
            title (string): Denotes the name of the profile and if populated will make the session persistent.
            headless (bool, optional): Whether the WebDriver should run in headless mode. Defaults to True.
            docker (bool, optional): Whether the session is running in a Docker container. Defaults to False.
        """
        self.title: str = title
        self.headless: bool = headless
        self.docker: bool = docker
        self.profile_path: str = ""
        self.external_code: bool = external_code
        self.need_code: bool = False
        self.driver = self.get_driver()

    def get_driver(self):
        """
        Gets the correct WebDriver for the operating system and initializes it with the necessary options.

        This method configures a Chromium WebDriver with various options to disable infobars, notifications, and GPU usage,
        and to enable certain experimental options. If the session is persistent, it also sets the user data directory.
        If the session is running in a Docker container, it configures the WebDriver accordingly.

        Returns:
            driver (selenium.webdriver.Chrome): The configured WebDriver instance.

        Raises:
            Exception: If there is an error initializing the WebDriver.
        """
        try:
            options = webdriver.ChromeOptions()
            options.add_argument("--disable-infobars")
            options.add_argument("--disable-notifications")
            if self.headless:
                options.add_argument("--headless")
                options.add_argument("--window-size=1920,1080")
                options.add_argument(
                    "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.3"
                )
            if self.title is not None:
                root = os.path.abspath(os.path.dirname(__file__))
                self.profile_path = os.path.join(root, "Profile", f"Chase_{self.title}")
                options.add_argument("user-data-dir=%s" % self.profile_path)
            options.add_argument("--disable-gpu")
            options.add_experimental_option("useAutomationExtension", False)
            options.add_experimental_option("excludeSwitches", ["enable-automation"])
            options.add_argument("disable-blink-features=AutomationControlled")
            options.add_argument("--log-level=3")
            if self.docker:
                options.add_argument("--no-sandbox")
                chrome_service = ChromiumService("/usr/bin/chromedriver")
                driver = webdriver.Chrome(service=chrome_service, options=options)
            else:
                driver = webdriver.Chrome(
                    service=ChromiumService(
                        ChromeDriverManager(chrome_type=ChromeType.CHROMIUM).install()
                    ),
                    options=options,
                )
        except Exception as e:
            print("Error getting Driver: \n")
            traceback.print_exc(e)
        if not self.headless:
            driver.maximize_window()
        else:
            driver.set_window_size(1920, 1080)
        return driver

    def get_login_code(self, queue):
        """
        Gets the login code from the user. Either from discord or from the terminal.
        
        Args:
            external_code (bool, optional): Whether the code should be retrieved externally. Defaults to False.
        
        Returns:
            code (str): The login code.
        """
        if not self.external_code:
            code = input("Please enter the code sent to your phone: ")
        else:
            self.need_code = True
            queue.put(self.need_code)
            event_handler = FileChange(".code")
            observer = Observer()
            observer.schedule(event_handler, path='.', recursive=False)
            observer.start()

            # Wait for the file to be modified
            for i in range(0, 120):
                sleep(1)
                if event_handler.file_modified:
                    break
                if i == 119:
                    raise Exception("Code not received in time cannot login.")

            observer.stop()
            observer.join()

            with open(".code", "r") as f:
                code = f.read()
        return code
    
    def login(self, username, password, last_four, queue):
        """
        Logs into the website with the provided username and password.

        This method navigates to the login page, enters the provided username and password into the appropriate fields,
        and submits the form. If the login is successful, the WebDriver will be redirected to the user's account page.

        Args:
            username (str): The user's username.
            password (str): The user's password.

        Raises:
            Exception: If there is an error during the login process.
        """
        try:
            self.driver.get(url=login_page())
            WebDriverWait(self.driver, 60).until(
                EC.presence_of_element_located((By.ID, "signin-button"))
            )
            self.driver.find_element(By.ID, "userId-text-input-field").send_keys(
                username
            )
            self.driver.find_element(By.ID, "password-text-input-field").send_keys(
                password
            )
            self.driver.find_element(By.ID, "signin-button").click()
            try:
                WebDriverWait(self.driver, 10).until(
                    EC.presence_of_element_located(
                        (
                            By.CSS_SELECTOR,
                            "#header-simplerAuth-dropdownoptions-styledselect",
                        )
                    )
                )
                contact_btn = self.driver.find_element(
                    By.CSS_SELECTOR, "#header-simplerAuth-dropdownoptions-styledselect"
                )
                contact_btn.click()
                options_ls = self.driver.find_elements(
                    By.CSS_SELECTOR, 'li[role="presentation"]'
                )
                for item in options_ls:
                    if "CALL_ME" not in item.text and str(last_four) in item.text:
                        item.click()
                        self.driver.find_element(
                            By.CSS_SELECTOR, 'button[type="submit"]'
                        ).click()

                WebDriverWait(self.driver, 60).until(EC.url_matches(auth_code_page()))
                code = self.get_login_code()
                self.driver.find_element(By.ID, "otpcode_input-input-field").send_keys(
                    code
                )
                self.driver.find_element(By.ID, "password_input-input-field").send_keys(
                    password
                )
                self.driver.find_element(
                    By.CSS_SELECTOR, 'button[type="submit"]'
                ).click()
                sleep(5)
            except TimeoutException:
                pass
            for _ in range(3):
                try:
                    self.driver.find_element(By.ID, "signin-button")
                    self.driver.refresh()
                    sleep(5)
                except NoSuchElementException:
                    queue.put(True)
                    return True
            raise Exception("Failed to login to Chase")
        except Exception as e:
            traceback.print_exc()
            print(f"Error logging into Chase: {e}")
            queue.put(False)
            return False

    def __getattr__(self, name):
        """
        Forwards unknown attribute access to session object.

        Args:
            name (str): The name of the attribute to be accessed.

        Returns:
            The value of the requested attribute from the session object.
        """
        return getattr(self.session, name)



import gzip
import json

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .session import ChaseSession
from .urls import account_info, landing_page


class AllAccount:
    """
    A class to manage all accounts associated with a ChaseSession.

    This class provides methods to retrieve and manage information about all accounts associated with a given session.

    Attributes:
        session (ChaseSession): The session associated with the accounts.
        all_account_info (dict): Information about all accounts associated with the session.
        account_connectors (dict): Connectors for all accounts associated with the session.
        total_value (float): The total value of all accounts.
        total_value_change (float): The change in total value of all accounts.

    Methods:
        get_all_account_info(): Retrieves information about all accounts associated with the session.
        get_account_connectors(): Retrieves connectors for all accounts associated with the session.
    """

    def __init__(self, session: ChaseSession):
        """
        Initializes an AllAccount object with a given ChaseSession.

        This method initializes an AllAccount object, sets the session attribute to the given ChaseSession, and retrieves the account information and connectors for all accounts associated with the session.

        Args:
            session (ChaseSession): The session associated with the accounts.
        """
        self.session = session
        self.all_account_info = self.get_all_account_info()
        self.account_connectors = self.get_account_connectors()

    def get_all_account_info(self):
        """
        Retrieves and returns information about all accounts associated with the session.

        This method navigates to the landing page, waits for the account information to load, and then retrieves the account information from the page.

        Returns:
            dict: A dictionary containing the account information, or None if the information could not be retrieved.
        """
        try:
            self.session.driver.get(landing_page())
            WebDriverWait(self.session.driver, 60).until(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR, "#INV_ACCOUNTS"))
            )
            for request in self.session.driver.requests:
                if request.response and request.url in account_info():
                    body = request.response.body
                    body = gzip.decompress(body).decode("utf-8")
                    account_json = json.loads(body)
                    for info in account_json["cache"]:
                        if (
                            info["url"]
                            == "/svc/rr/accounts/secure/v1/account/detail/inv/list"
                        ):
                            invest_json = info["response"]["chaseInvestments"]
                    if request.response.status_code == 200:
                        self.total_value = invest_json["investmentSummary"][
                            "accountValue"
                        ]
                        self.total_value_change = invest_json["investmentSummary"][
                            "accountValueChange"
                        ]
                        return invest_json
                    return None
        except (TimeoutException, NoSuchElementException):
            print("Timed out waiting for page to load")
            return None

    def get_account_connectors(self):
        """
        Retrieves and returns connectors for all accounts associated with the session.

        This method iterates over all accounts in the all_account_info attribute and creates a dictionary where the keys are account IDs and the values are lists containing the corresponding account masks.

        Returns:
            dict: A dictionary containing the account connectors, or None if the all_account_info attribute is None.
        """
        if self.all_account_info is None:
            return None
        account_dict = {}
        for item in self.all_account_info["accounts"]:
            account_dict[item["accountId"]] = [item["mask"]]
        return account_dict


class AccountDetails:
    """
    A class to manage the details of a specific account associated with a ChaseSession.

    This class provides methods to retrieve and manage information about a specific account associated with a given session.

    Attributes:
        all_account_info (dict): Information about all accounts associated with the session.
        account_id (str): The ID of the account.
        nickname (str): The nickname of the account.
        mask (str): The mask of the account.
        detail_type (str): The detail type of the account.
        account_value (float): The value of the account.
        account_value_change (float): The change in value of the account.
        eda (str): The EDA of the account.
        ira (bool): Whether the account is an IRA.
        view_balance (bool): Whether the balance of the account can be viewed.
        prior_year_ira (bool): Whether the account is a prior year IRA.
        show_xfer (bool): Whether the account shows transfers.

    Methods:
        get_account_details(): Retrieves and sets the details of the account.
    """

    def __init__(self, account_id, all_account: AllAccount):
        """
        Initializes an AccountDetails object with a given account ID and AllAccount object.

        Args:
            account_id (str): The ID of the account.
            all_account (AllAccount): The AllAccount object associated with the session.
        """
        self.all_account_info = all_account.all_account_info
        self.account_id: str = account_id
        self.nickname: str = None
        self.mask: str = None
        self.detail_type: str = None
        self.account_value: float = None
        self.account_value_change: float = None
        self.eda: str = None
        self.ira: bool = None
        self.view_balance: bool = None
        self.prior_year_ira: bool = None
        self.show_xfer: bool = None
        self.get_account_details()

    def get_account_details(self):
        """
        Retrieves and sets the details of the account.

        This method iterates over all accounts in the all_account_info attribute and sets the attributes of the AccountDetails object to the details of the account with the matching account ID.

        Returns:
            None
        """
        for item in self.all_account_info["accounts"]:
            if item["accountId"] == self.account_id:
                self.mask = item["mask"]
                self.nickname = item["nickname"]
                self.detail_type = item["detailType"]
                self.account_value = float(item["accountValue"])
                self.account_value_change = float(item["accountValueChange"])
                self.eda = bool(item["eda"])
                self.ira = bool(item["ira"])
                self.view_balance = bool(item["viewBalance"])
                self.prior_year_ira = bool(item["priorYearIra"])
                self.show_xfer = bool(item["showXfer"])
